/* Copyright (c) 2005, Dmitry Xmelkov
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.
   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE. */

/* $Id$ */

/* Algorithm of division by 10 is get from:
     Donald E. Knuth "The art of computer programming"
     Volume 2, page 371 (in Russian).	*/

#include "macros.inc"
#include "xtoa_fast.h"

/* --------------------------------------------------------------------
   char * itoa_fast (int val, char * str, int base)
   char * utoa_fast (unsigned val, char * str, int base)
   char * itoa_width (int val, char * str, int base, size_t width)
   char * utoa_width (unsigned val, char * str, int base, size_t width)
*/
#define	XTOA_MINUS	0x8000		/* 'minus' is present	*/
#define	XTOA_DECIMAL	0x4000		/* base is 10		*/
#define	XTOA_HEX	0x2000		/* base is 16		*/
#define	XTOA_WIDTH	0x1000		/* 'width' is present	*/

    /* Input	*/
#define	val_lo	r24
#define	val_hi	r25
#define	str_lo	r22
#define	str_hi	r23
#define	base	r20
#define	flags	r21	/* high byte of `base'	*/
#define	width_lo r18
#define	width_hi r19

    /* Intermediate	*/
#define	val_hlo	r26	/* val: bits 23..16		*/
#define	tmp_lo	r22	/* temporary for shifted `val'	*/
#define	tmp_hi	r23
#define	cnt	r27	/* shift loop counter		*/
#define	symb	r20	/* write to string		*/
#define	len_lo	r26	/* string length, for filling	*/
#define	len_hi	r27

    /* Fixed	*/
#define	zero	r1

	.text
	
	.global	itoa_width
	.type	itoa_width, "function"
itoa_width:
	andi	flags, (XTOA_UPPER | XTOA_PREFIX) >> 8
	ori	flags, XTOA_WIDTH >> 8
	rjmp	.L_itoa
	.size	itoa_width, . - itoa_width

	.global	utoa_width
	.type	utoa_width, "function"
utoa_width:
	andi	flags, (XTOA_UPPER | XTOA_PREFIX) >> 8
	ori	flags, XTOA_WIDTH >> 8
	rjmp	.L_utoa
	.size	utoa_width, . - utoa_width

	.global	utoa_fast
	.type	utoa_fast, "function"
utoa_fast:
	andi	flags, (XTOA_UPPER | XTOA_PREFIX) >> 8
	rjmp	.L_utoa
	.size	utoa_fast, . - utoa_fast
	
	.global	itoa_fast
	.type	itoa_fast, "function"
itoa_fast:
	andi	flags, (XTOA_UPPER | XTOA_PREFIX) >> 8
.L_itoa:
	cpi	base, 10
	brne	.L_utoa		; sign is only for decimals
  ; sign ?
	tst	val_hi
	brpl	.L_utoa
  ; minus
	com	val_hi
	neg	val_lo
	sbci	val_hi, hi8(-1)
	ori	flags, XTOA_MINUS >> 8
.L_utoa:
  ; prologue
	X_movw	ZL, str_lo
	push	str_hi
	push	str_lo
	clr	val_hlo
  ; base ?
	cpi	base, 10
	breq	.L_dec
	cpi	base, 16
	breq	.L_hex
	cpi	base, 8
	breq	.L_oct
	cpi	base, 2
	breq	.L_bin
  ; unknown base --> no conversion
	clr	flags
	rjmp	.L_pop_str

  ; Binary
.L_bin:
  ; lsr `val'
	lsr	val_hi
	ror	val_lo
  ; output '0' + carry
	ldi	symb, '0'
	adc	symb, zero
	st	Z+, symb
  ; tst `val'
	adiw	val_lo, 0
	brne	.L_bin
  ; epilogue
	andi	flags, XTOA_WIDTH >> 8
	rjmp	.L_pop_str

  ; Octal
.L_oct:
  ; output 3-bit digit
	mov	symb, val_lo
	andi	symb, 7
	subi	symb, lo8(-'0')
	st	Z+, symb
  ; lsr `val'
	ldi	cnt, 3
	rcall	.L_lsr
	brne	.L_oct
  ; epilogue
	rjmp	.L_pop_str

  ; Hexidecimal
.L_hex:
  ; output 4-bit digit
	mov	symb, val_lo
	andi	symb, 0x0f
	subi	symb, lo8(-'0')
	cpi	symb, '9' + 1
	brlo	1f
	subi	symb, lo8('9' + 1 - 'a')
	sbrc	flags, ntz(XTOA_UPPER) - 8
	subi	symb, lo8('a' - 'A')
1:	st	Z+, symb
  ; lsr `val'
	rcall	.L_lsr_4
	brne	.L_hex
  ; epilogue
	ori	flags, XTOA_HEX >> 8
	rjmp	.L_pop_str

  ; Decimal
.L_dec:
	push	val_lo			; to calculate remainder
  ; val &= ~1
	andi	val_lo, ~1
  ; val += 2
	subi	val_lo, lo8(-2)
	sbci	val_hi, hi8(-2)
	sbci	val_hlo, hlo8(-2)
  ; val += val/2
	ldi	cnt, 1
	rcall	.L_div_add
  ; val += val/16
	ldi	cnt, 4
	rcall	.L_div_add
  ; val += val/256
	add	val_lo, val_hi
	adc	val_hi, val_hlo
	adc	val_hlo, zero
  ; val += val/65536
	add	val_lo, val_hlo
	adc 	val_hi, zero
	adc	val_hlo, zero
  ; division result: val /= 16
	rcall	.L_lsr_4		; val_hlo become 0
  ; rem:  val_original - 10*val
	pop	tmp_hi
#if  defined(__AVR_ENHANCED__) && __AVR_ENHANCED__
	ldi	tmp_lo, 10
	mul	tmp_lo, val_lo
	clr	r1
#else
	mov	r0, val_lo
	lsl	r0
	sub	tmp_hi, r0
	lsl	r0
	lsl	r0
#endif
	sub	tmp_hi, r0
  ; output digit
	subi	tmp_hi, lo8(-'0')
	st	Z+, tmp_hi
  ; quotient == 0 ?
	sbiw	val_lo, 0
	brne	.L_dec
  ; minus, if needed
	sbrs	flags, ntz(XTOA_MINUS) - 8
	rjmp	3f
	ldi	symb, '-'
	st	Z+, symb
  ; for common epilogue
3:	andi	flags, XTOA_WIDTH >> 8
	ori	flags, XTOA_DECIMAL >> 8
  ; string address restore
.L_pop_str:
	pop	r24		; str_lo
	pop	r25		; str_hi
  ; length calculation
	X_movw	len_lo, ZL
	sub	len_lo, r24
	sbc	len_hi, r25
  ; is width used?
	sbrs	flags, ntz(XTOA_WIDTH) - 8
	rjmp	.L_prefix
  ; is width too short?
	cpi	width_lo, lo8(2)
	cpc	width_hi, zero
	brlo	.L_prefix	; width <2 is too short for any base
	brne	4f		; width >2 is normal
	sbrs	flags, ntz(XTOA_HEX) - 8
	rjmp	4f
	sbrc	flags, ntz(XTOA_PREFIX) - 8
	rjmp	.L_prefix	; width 2 is too short for hex with prefix
  ; length += strlen(prefix)	(and skip test 'val == 0' in prefix output)
4:	sbrs	flags, ntz(XTOA_PREFIX) - 8
	rjmp	5f
	adiw	len_lo, 1
	sbrc	flags, ntz(XTOA_HEX) - 8
	adiw	len_lo, 1	; prefix for hex is 2 bytes long
  ; filling
5:	ldi	symb, '0'
	sbrc	flags, ntz(XTOA_DECIMAL) - 8
	ldi	symb, ' '
6:	cp	len_lo, width_lo
	cpc	len_hi, width_hi
	brsh	.L_prefix
	st	Z+, symb
	adiw	len_lo, 1
	rjmp	6b
  ; prefix?
.L_prefix:
	sbrs	flags, ntz(XTOA_PREFIX) - 8
	rjmp	.L_reverse	; not needed
  ; val == 0 ?
	cpi	len_lo, lo8(2)		; Length checking is needed
	cpc	len_hi, zero		;   in the case of `width'.
	brsh	7f
	ld	symb, -Z
	adiw	ZL, 1
	cpi	symb, '0'
	breq	.L_reverse	; no prefix for '0'
  ; prefix output
7:	sbrs	flags, ntz(XTOA_HEX) - 8
	rjmp	8f
	ldi	symb, 'x'
	sbrc	flags, ntz(XTOA_UPPER) - 8
	ldi	symb, 'X'
	st	Z+, symb
8:	ldi	symb, '0'
	st	Z+, symb
  ; reverse string
.L_reverse:
	st	Z, zero
	XJMP	strrev

  ; shift 3-bytes val and test result by 0
.L_lsr_4:
	ldi	cnt, 4
.L_lsr:
	lsr	val_hlo
	ror	val_hi
	ror	val_lo
	dec	cnt
	brne	.L_lsr
  ; tst
	sbiw	val_lo, 0
	ret

  ; val += val >> cnt
.L_div_add:
  ; copy to temporary
	X_movw	tmp_lo, val_lo
	mov	r0, val_hlo
  ; lsr	temporary
2:	lsr	r0
	ror	tmp_hi
	ror	tmp_lo
	dec	cnt
	brne	2b
  ; add
	add	val_lo, tmp_lo
	adc	val_hi, tmp_hi
	adc	val_hlo, r0
	ret

	.size	itoa_fast, . - itoa_fast
	.end
